"""Tests for WebFetchTool."""

from __future__ import annotations

from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from hauba.tools.fetch import WebFetchTool


@pytest.fixture
def fetch_tool() -> WebFetchTool:
    return WebFetchTool()


class TestWebFetchTool:
    """Test WebFetchTool.execute()."""

    def test_name_and_description(self, fetch_tool: WebFetchTool) -> None:
        assert fetch_tool.name == "web_fetch"
        assert "fetch" in fetch_tool.description.lower()

    @pytest.mark.asyncio
    async def test_missing_url(self, fetch_tool: WebFetchTool) -> None:
        result = await fetch_tool.execute()
        assert not result.success
        assert "URL required" in result.error

    @pytest.mark.asyncio
    async def test_successful_html_fetch(self, fetch_tool: WebFetchTool) -> None:
        fake_html = (
            "<html><head><title>Test Page</title></head><body><p>Hello world</p></body></html>"
        )
        mock_resp = MagicMock()
        mock_resp.text = fake_html
        mock_resp.headers = {"content-type": "text/html"}
        mock_resp.raise_for_status = MagicMock()

        mock_client = AsyncMock()
        mock_client.__aenter__ = AsyncMock(return_value=mock_client)
        mock_client.__aexit__ = AsyncMock(return_value=False)
        mock_client.get = AsyncMock(return_value=mock_resp)

        with patch("hauba.tools.fetch.httpx.AsyncClient", return_value=mock_client):
            result = await fetch_tool.execute(url="https://example.com")

        assert result.success
        assert "Hello world" in result.output

    @pytest.mark.asyncio
    async def test_json_response(self, fetch_tool: WebFetchTool) -> None:
        mock_resp = MagicMock()
        mock_resp.text = '{"key": "value"}'
        mock_resp.headers = {"content-type": "application/json"}
        mock_resp.raise_for_status = MagicMock()

        mock_client = AsyncMock()
        mock_client.__aenter__ = AsyncMock(return_value=mock_client)
        mock_client.__aexit__ = AsyncMock(return_value=False)
        mock_client.get = AsyncMock(return_value=mock_resp)

        with patch("hauba.tools.fetch.httpx.AsyncClient", return_value=mock_client):
            result = await fetch_tool.execute(url="https://api.example.com/data")

        assert result.success
        assert '"key"' in result.output

    @pytest.mark.asyncio
    async def test_timeout_error(self, fetch_tool: WebFetchTool) -> None:
        import httpx

        mock_client = AsyncMock()
        mock_client.__aenter__ = AsyncMock(return_value=mock_client)
        mock_client.__aexit__ = AsyncMock(return_value=False)
        mock_client.get = AsyncMock(side_effect=httpx.TimeoutException("timeout"))

        with patch("hauba.tools.fetch.httpx.AsyncClient", return_value=mock_client):
            result = await fetch_tool.execute(url="https://slow.example.com")

        assert not result.success
        assert "Timeout" in result.error

    @pytest.mark.asyncio
    async def test_http_error(self, fetch_tool: WebFetchTool) -> None:
        import httpx

        mock_resp = MagicMock()
        mock_resp.status_code = 404
        mock_resp.raise_for_status = MagicMock(
            side_effect=httpx.HTTPStatusError("Not Found", request=MagicMock(), response=mock_resp)
        )

        mock_client = AsyncMock()
        mock_client.__aenter__ = AsyncMock(return_value=mock_client)
        mock_client.__aexit__ = AsyncMock(return_value=False)
        mock_client.get = AsyncMock(return_value=mock_resp)

        with patch("hauba.tools.fetch.httpx.AsyncClient", return_value=mock_client):
            result = await fetch_tool.execute(url="https://example.com/404")

        assert not result.success
        assert "404" in result.error


class TestHtmlToText:
    """Test WebFetchTool._html_to_text()."""

    def test_extract_title(self) -> None:
        html = "<html><head><title>My Title</title></head><body>Content</body></html>"
        title, _text = WebFetchTool._html_to_text(html)
        assert title == "My Title"

    def test_strip_scripts_and_styles(self) -> None:
        html = (
            "<html><body>"
            "<script>alert('bad')</script>"
            "<style>body{color:red}</style>"
            "<p>Visible content</p>"
            "</body></html>"
        )
        _title, text = WebFetchTool._html_to_text(html)
        assert "alert" not in text
        assert "color:red" not in text
        assert "Visible content" in text

    def test_empty_html(self) -> None:
        title, _text = WebFetchTool._html_to_text("")
        assert title == ""


class TestParametersSchema:
    """Test tool schema."""

    def test_schema_has_url(self) -> None:
        tool = WebFetchTool()
        schema = tool._parameters_schema()
        assert "url" in schema["properties"]
        assert "url" in schema["required"]



"""End-to-end Playwright browser tests — real browser, real interactions.

These tests spin up a local HTTP server with a mini web app and use
Hauba's BrowserTool to navigate, click, type, extract, and screenshot.

Run with:
    pytest tests/e2e/test_browser_demo.py -v --headed   # visible browser
    pytest tests/e2e/test_browser_demo.py -v             # headless (CI)

Mark: requires playwright chromium to be installed.
"""

from __future__ import annotations

import threading
from collections.abc import Generator
from http.server import HTTPServer, SimpleHTTPRequestHandler
from pathlib import Path

import pytest

from hauba.tools.browser import PLAYWRIGHT_AVAILABLE, BrowserTool

# Skip entire module if Playwright is not installed
pytestmark = pytest.mark.skipif(not PLAYWRIGHT_AVAILABLE, reason="Playwright not installed")

# ---------- Local test web app ----------

_APP_HTML = """\
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <title>Hauba Demo App</title>
    <style>
        * { box-sizing: border-box; margin: 0; padding: 0; }
        body { font-family: 'Segoe UI', sans-serif; background: #0f172a; color: #e2e8f0; }
        .container { max-width: 800px; margin: 0 auto; padding: 40px 20px; }
        h1 { font-size: 2.5rem; margin-bottom: 8px; background: linear-gradient(135deg, #38bdf8, #818cf8);
             -webkit-background-clip: text; -webkit-text-fill-color: transparent; }
        .subtitle { color: #94a3b8; font-size: 1.1rem; margin-bottom: 32px; }
        .card { background: #1e293b; border-radius: 12px; padding: 24px; margin-bottom: 20px;
                border: 1px solid #334155; }
        .card h2 { color: #38bdf8; margin-bottom: 12px; font-size: 1.3rem; }
        input[type="text"] { width: 100%; padding: 12px 16px; border-radius: 8px; border: 1px solid #475569;
                             background: #0f172a; color: #e2e8f0; font-size: 1rem; margin-bottom: 12px; }
        button { padding: 12px 24px; border-radius: 8px; border: none; cursor: pointer;
                 font-size: 1rem; font-weight: 600; transition: all 0.2s; }
        .btn-primary { background: #3b82f6; color: white; }
        .btn-primary:hover { background: #2563eb; }
        .btn-success { background: #22c55e; color: white; }
        .btn-danger { background: #ef4444; color: white; margin-left: 8px; }
        #task-list { list-style: none; margin-top: 16px; }
        #task-list li { padding: 12px 16px; background: #0f172a; border-radius: 8px;
                        margin-bottom: 8px; display: flex; justify-content: space-between;
                        align-items: center; border: 1px solid #334155; }
        #task-list li .task-text { flex: 1; }
        #task-list li.completed .task-text { text-decoration: line-through; color: #64748b; }
        .status-bar { margin-top: 24px; padding: 16px; background: #1e293b;
                      border-radius: 8px; text-align: center; border: 1px solid #334155; }
        #status { color: #38bdf8; font-weight: 600; }
        #task-count { color: #94a3b8; font-size: 0.9rem; margin-top: 8px; }
        .stats { display: flex; gap: 20px; margin-top: 20px; }
        .stat { flex: 1; text-align: center; padding: 16px; background: #1e293b;
                border-radius: 8px; border: 1px solid #334155; }
        .stat-value { font-size: 2rem; font-weight: 700; color: #38bdf8; }
        .stat-label { font-size: 0.85rem; color: #94a3b8; margin-top: 4px; }
    </style>
</head>
<body>
    <div class="container">
        <h1>Hauba AI Workstation</h1>
        <p class="subtitle">Task Manager Demo &mdash; Powered by Playwright</p>

        <div class="card">
            <h2>Add Task</h2>
            <input type="text" id="task-input" placeholder="What needs to be done?" />
            <button class="btn-primary" id="add-btn" onclick="addTask()">Add Task</button>
        </div>

        <div class="card">
            <h2>Tasks</h2>
            <ul id="task-list"></ul>
            <div id="empty-msg" style="color: #64748b; text-align: center; padding: 20px;">
                No tasks yet. Add one above!
            </div>
        </div>

        <div class="stats">
            <div class="stat">
                <div class="stat-value" id="total-count">0</div>
                <div class="stat-label">Total</div>
            </div>
            <div class="stat">
                <div class="stat-value" id="active-count">0</div>
                <div class="stat-label">Active</div>
            </div>
            <div class="stat">
                <div class="stat-value" id="done-count">0</div>
                <div class="stat-label">Completed</div>
            </div>
        </div>

        <div class="status-bar">
            <div id="status">Ready</div>
            <div id="task-count">0 tasks</div>
        </div>
    </div>

    <script>
        let tasks = [];

        function addTask() {
            const input = document.getElementById('task-input');
            const text = input.value.trim();
            if (!text) return;

            tasks.push({ text, completed: false, id: Date.now() });
            input.value = '';
            render();
            updateStatus('Task added: ' + text);
        }

        function toggleTask(id) {
            const task = tasks.find(t => t.id === id);
            if (task) {
                task.completed = !task.completed;
                render();
                updateStatus(task.completed ? 'Task completed!' : 'Task reopened');
            }
        }

        function deleteTask(id) {
            tasks = tasks.filter(t => t.id !== id);
            render();
            updateStatus('Task deleted');
        }

        function render() {
            const list = document.getElementById('task-list');
            const empty = document.getElementById('empty-msg');

            if (tasks.length === 0) {
                list.innerHTML = '';
                empty.style.display = 'block';
            } else {
                empty.style.display = 'none';
                list.innerHTML = tasks.map(t => `
                    <li class="${t.completed ? 'completed' : ''}">
                        <span class="task-text">${t.text}</span>
                        <div>
                            <button class="btn-success" onclick="toggleTask(${t.id})">
                                ${t.completed ? 'Undo' : 'Done'}
                            </button>
                            <button class="btn-danger" onclick="deleteTask(${t.id})">Del</button>
                        </div>
                    </li>
                `).join('');
            }

            const total = tasks.length;
            const done = tasks.filter(t => t.completed).length;
            const active = total - done;

            document.getElementById('total-count').textContent = total;
            document.getElementById('active-count').textContent = active;
            document.getElementById('done-count').textContent = done;
            document.getElementById('task-count').textContent = `${total} task${total !== 1 ? 's' : ''}`;
        }

        function updateStatus(msg) {
            document.getElementById('status').textContent = msg;
        }

        // Handle Enter key
        document.getElementById('task-input').addEventListener('keypress', function(e) {
            if (e.key === 'Enter') addTask();
        });
    </script>
</body>
</html>
"""


class _DemoHandler(SimpleHTTPRequestHandler):
    """Serve the demo HTML from memory."""

    def do_GET(self) -> None:
        self.send_response(200)
        self.send_header("Content-Type", "text/html; charset=utf-8")
        self.end_headers()
        self.wfile.write(_APP_HTML.encode("utf-8"))

    def log_message(self, format: str, *args: object) -> None:
        pass  # suppress server logs during tests


@pytest.fixture(scope="module")
def demo_server() -> Generator[str, None, None]:
    """Start a local HTTP server with the demo app."""
    server = HTTPServer(("127.0.0.1", 0), _DemoHandler)
    port = server.server_address[1]
    thread = threading.Thread(target=server.serve_forever, daemon=True)
    thread.start()
    url = f"http://127.0.0.1:{port}"
    yield url
    server.shutdown()


@pytest.fixture
def browser_tool(request: pytest.FixtureRequest) -> BrowserTool:
    """Create a BrowserTool. Use --headed flag to see the browser."""
    headed = request.config.getoption("--headed", default=False)
    return BrowserTool(
        session_name="e2e-test",
        headless=not headed,
        stealth=False,
    )


def pytest_addoption(parser: pytest.Parser) -> None:
    """Add --headed flag for visible browser tests."""
    parser.addoption(
        "--headed",
        action="store_true",
        default=False,
        help="Run browser tests in headed mode (visible browser window)",
    )


# ---------- Tests ----------


class TestBrowserNavigate:
    """Test navigating to the demo app."""

    async def test_navigate_to_demo(self, browser_tool: BrowserTool, demo_server: str) -> None:
        """Navigate to the demo app and verify the page loads."""
        try:
            result = await browser_tool.execute(action="navigate", url=demo_server)
            assert result.success, f"Navigate failed: {result.error}"
            assert "Navigated to" in result.output
        finally:
            await browser_tool.close()

    async def test_extract_page_title(self, browser_tool: BrowserTool, demo_server: str) -> None:
        """Extract the page heading text."""
        try:
            await browser_tool.execute(action="navigate", url=demo_server)
            result = await browser_tool.execute(action="extract", selector="h1")
            assert result.success
            assert "Hauba AI Workstation" in result.output
        finally:
            await browser_tool.close()

    async def test_extract_subtitle(self, browser_tool: BrowserTool, demo_server: str) -> None:
        """Extract the subtitle text."""
        try:
            await browser_tool.execute(action="navigate", url=demo_server)
            result = await browser_tool.execute(action="extract", selector=".subtitle")
            assert result.success
            assert "Playwright" in result.output
        finally:
            await browser_tool.close()


class TestBrowserInteraction:
    """Test clicking, typing, and form interactions."""

    async def test_type_into_input(self, browser_tool: BrowserTool, demo_server: str) -> None:
        """Type a task name into the input field."""
        try:
            await browser_tool.execute(action="navigate", url=demo_server)
            result = await browser_tool.execute(
                action="type", selector="#task-input", text="Build a REST API"
            )
            assert result.success
            assert "Typed" in result.output
        finally:
            await browser_tool.close()

    async def test_add_task_via_button(self, browser_tool: BrowserTool, demo_server: str) -> None:
        """Type a task and click the Add button."""
        try:
            await browser_tool.execute(action="navigate", url=demo_server)
            await browser_tool.execute(
                action="type", selector="#task-input", text="Deploy to production"
            )
            result = await browser_tool.execute(action="click", selector="#add-btn")
            assert result.success

            # Verify the task appears in the list
            extract = await browser_tool.execute(action="extract", selector="#task-list")
            assert extract.success
            assert "Deploy to production" in extract.output
        finally:
            await browser_tool.close()

    async def test_add_multiple_tasks(self, browser_tool: BrowserTool, demo_server: str) -> None:
        """Add multiple tasks and verify the counter updates."""
        try:
            await browser_tool.execute(action="navigate", url=demo_server)

            tasks = ["Write unit tests", "Set up CI/CD", "Update README"]
            for task in tasks:
                await browser_tool.execute(action="type", selector="#task-input", text=task)
                await browser_tool.execute(action="click", selector="#add-btn")

            # Check total count
            count = await browser_tool.execute(action="extract", selector="#total-count")
            assert count.success
            assert "3" in count.output

            # Check all tasks are in the list
            list_text = await browser_tool.execute(action="extract", selector="#task-list")
            for task in tasks:
                assert task in list_text.output
        finally:
            await browser_tool.close()

    async def test_complete_task(self, browser_tool: BrowserTool, demo_server: str) -> None:
        """Add a task, then mark it as done."""
        try:
            await browser_tool.execute(action="navigate", url=demo_server)

            # Add a task
            await browser_tool.execute(action="type", selector="#task-input", text="Fix login bug")
            await browser_tool.execute(action="click", selector="#add-btn")

            # Click "Done" button on the task
            result = await browser_tool.execute(action="click", selector="#task-list .btn-success")
            assert result.success

            # Verify done count updated
            done = await browser_tool.execute(action="extract", selector="#done-count")
            assert done.success
            assert "1" in done.output

            # Verify status message
            status = await browser_tool.execute(action="extract", selector="#status")
            assert status.success
            assert "completed" in status.output.lower()
        finally:
            await browser_tool.close()


class TestBrowserScreenshot:
    """Test screenshot capability."""

    async def test_screenshot_demo_app(
        self, browser_tool: BrowserTool, demo_server: str, tmp_path: Path
    ) -> None:
        """Take a screenshot of the demo app."""
        try:
            await browser_tool.execute(action="navigate", url=demo_server)

            # Add some tasks for a nice screenshot
            for task in ["Build API", "Write Tests", "Deploy"]:
                await browser_tool.execute(action="type", selector="#task-input", text=task)
                await browser_tool.execute(action="click", selector="#add-btn")

            # Take screenshot
            shot_path = str(tmp_path / "hauba_demo.png")
            result = await browser_tool.execute(action="screenshot", path=shot_path)
            assert result.success
            assert Path(shot_path).exists()
            assert Path(shot_path).stat().st_size > 1000  # should be a real image
        finally:
            await browser_tool.close()


class TestBrowserAdvanced:
    """Test advanced browser features — scroll, evaluate, wait."""

    async def test_evaluate_javascript(self, browser_tool: BrowserTool, demo_server: str) -> None:
        """Execute JavaScript in the page context."""
        try:
            await browser_tool.execute(action="navigate", url=demo_server)
            result = await browser_tool.execute(
                action="evaluate", script="document.title || 'Hauba Demo App'"
            )
            assert result.success
            assert "Hauba" in result.output
        finally:
            await browser_tool.close()

    async def test_scroll_page(self, browser_tool: BrowserTool, demo_server: str) -> None:
        """Scroll the page down and back up."""
        try:
            await browser_tool.execute(action="navigate", url=demo_server)

            down = await browser_tool.execute(action="scroll", direction="down", amount=300)
            assert down.success
            assert "Scrolled down" in down.output

            up = await browser_tool.execute(action="scroll", direction="up", amount=300)
            assert up.success
            assert "Scrolled up" in up.output
        finally:
            await browser_tool.close()

    async def test_wait_for_element(self, browser_tool: BrowserTool, demo_server: str) -> None:
        """Wait for an element to be visible."""
        try:
            await browser_tool.execute(action="navigate", url=demo_server)
            result = await browser_tool.execute(action="wait", selector="h1", timeout=5000)
            assert result.success
            assert "visible" in result.output.lower()
        finally:
            await browser_tool.close()

    async def test_extract_full_page(self, browser_tool: BrowserTool, demo_server: str) -> None:
        """Extract all text content from the page."""
        try:
            await browser_tool.execute(action="navigate", url=demo_server)
            result = await browser_tool.execute(action="extract")
            assert result.success
            assert "Hauba AI Workstation" in result.output
            assert "Task Manager Demo" in result.output
        finally:
            await browser_tool.close()

    async def test_page_info(self, browser_tool: BrowserTool, demo_server: str) -> None:
        """Get current page URL and title."""
        try:
            await browser_tool.execute(action="navigate", url=demo_server)
            info = await browser_tool.get_page_info()
            assert demo_server in info["url"]
        finally:
            await browser_tool.close()


class TestBrowserFullWorkflow:
    """Full end-to-end workflow — like a real Hauba session."""

    async def test_full_task_management_workflow(
        self, browser_tool: BrowserTool, demo_server: str, tmp_path: Path
    ) -> None:
        """Complete workflow: navigate, add tasks, complete one, screenshot."""
        try:
            # 1. Navigate to the app
            nav = await browser_tool.execute(action="navigate", url=demo_server)
            assert nav.success

            # 2. Verify the page loaded
            title = await browser_tool.execute(action="extract", selector="h1")
            assert "Hauba AI Workstation" in title.output

            # 3. Add three tasks
            task_names = [
                "Design database schema",
                "Implement REST API endpoints",
                "Write integration tests",
            ]
            for name in task_names:
                await browser_tool.execute(action="type", selector="#task-input", text=name)
                await browser_tool.execute(action="click", selector="#add-btn")

            # 4. Verify all tasks appear
            tasks = await browser_tool.execute(action="extract", selector="#task-list")
            for name in task_names:
                assert name in tasks.output

            # 5. Verify stats
            total = await browser_tool.execute(action="extract", selector="#total-count")
            assert "3" in total.output

            active = await browser_tool.execute(action="extract", selector="#active-count")
            assert "3" in active.output

            # 6. Complete the first task
            await browser_tool.execute(
                action="click", selector="#task-list li:first-child .btn-success"
            )

            # 7. Verify done count
            done = await browser_tool.execute(action="extract", selector="#done-count")
            assert "1" in done.output

            # 8. Take final screenshot
            shot = str(tmp_path / "workflow_complete.png")
            result = await browser_tool.execute(action="screenshot", path=shot)
            assert result.success
            assert Path(shot).exists()

            # 9. Check status bar
            status = await browser_tool.execute(action="extract", selector="#status")
            assert status.success

        finally:
            await browser_tool.close()



"""Tests for BashTool."""

from __future__ import annotations

import pytest

from hauba.tools.bash import BashTool


@pytest.mark.asyncio
async def test_echo_command() -> None:
    tool = BashTool()
    result = await tool.execute(command="echo hello")
    assert result.success
    assert "hello" in result.output


@pytest.mark.asyncio
async def test_failing_command() -> None:
    tool = BashTool()
    result = await tool.execute(command="exit 1")
    assert not result.success
    assert result.exit_code == 1


@pytest.mark.asyncio
async def test_timeout() -> None:
    tool = BashTool(timeout=1)
    result = await tool.execute(command="sleep 10", timeout=1)
    assert not result.success
    assert "timed out" in result.error.lower()


@pytest.mark.asyncio
async def test_no_command() -> None:
    tool = BashTool()
    result = await tool.execute()
    assert not result.success
    assert "No command" in result.error



"""Tests for CopilotEngine — skill injection, configuration, and availability."""

from __future__ import annotations

from hauba.engine.copilot_engine import CopilotEngine
from hauba.engine.types import EngineConfig, ProviderType


def test_build_system_prompt_contains_protocol() -> None:
    """Default system prompt contains the execution protocol."""
    config = EngineConfig(provider=ProviderType.ANTHROPIC, api_key="test")
    engine = CopilotEngine(config)
    prompt = engine._build_hauba_system_prompt()
    assert "EXECUTION PROTOCOL" in prompt
    assert "Phase 1: UNDERSTAND" in prompt
    assert "Phase 3: IMPLEMENT" in prompt
    assert "Phase 4: VERIFY" in prompt


def test_build_system_prompt_with_skill_context() -> None:
    """Skill context is appended to the system prompt."""
    config = EngineConfig(provider=ProviderType.ANTHROPIC, api_key="test")
    engine = CopilotEngine(config, skill_context="## Test Skill\n- Build REST APIs")
    prompt = engine._build_hauba_system_prompt()
    assert "## Test Skill" in prompt
    assert "- Build REST APIs" in prompt


def test_build_system_prompt_without_skill_context() -> None:
    """Without skill context, the prompt is still valid."""
    config = EngineConfig(provider=ProviderType.ANTHROPIC, api_key="test")
    engine = CopilotEngine(config, skill_context="")
    prompt = engine._build_hauba_system_prompt()
    assert "Hauba AI Workstation" in prompt
    assert "## Test Skill" not in prompt


def test_engine_is_available_checks_import() -> None:
    """Engine reports availability based on copilot SDK import."""
    config = EngineConfig(provider=ProviderType.ANTHROPIC, api_key="test")
    engine = CopilotEngine(config)
    # Returns bool (True if SDK installed, False otherwise)
    assert isinstance(engine.is_available, bool)


def test_system_prompt_enforces_tool_use() -> None:
    """System prompt tells the agent to USE tools, not just describe."""
    config = EngineConfig(provider=ProviderType.ANTHROPIC, api_key="test")
    engine = CopilotEngine(config)
    prompt = engine._build_hauba_system_prompt()
    assert "MUST use your tools" in prompt
    assert "no placeholders" in prompt.lower() or "no TODOs" in prompt



"""Phase 4 Integration Tests — Channels + Notifier + Gateway pipeline."""

from __future__ import annotations

from unittest.mock import AsyncMock

from hauba.channels.gateway import ChannelGateway, ChannelMessage, ChannelType
from hauba.core.events import EventEmitter
from hauba.core.notifier import Notification, NotificationLevel, Notifier


async def test_full_notification_pipeline() -> None:
    """Events → Notifier → Sinks → Gateway → Channels.

    Verifies the complete Phase 4 pipeline: an event triggers a notification
    which is delivered through the gateway to registered channels.
    """
    events = EventEmitter()
    notifier = Notifier(events)

    # Track all notifications
    received: list[Notification] = []

    async def capture_sink(n: Notification) -> None:
        received.append(n)

    notifier.add_sink(capture_sink)
    notifier.subscribe()

    # Set up gateway with mock channel
    gateway = ChannelGateway(events)
    mock_channel = AsyncMock()
    mock_channel.broadcast = AsyncMock()
    gateway.register_channel(ChannelType.WEB, mock_channel)

    # Simulate a task lifecycle
    await events.emit("task.started", {"task_id": "t-1", "instruction": "build app"})
    await events.emit("milestone.completed", {"milestone_id": "m-1"}, task_id="t-1")
    await events.emit("task.completed", {"task_id": "t-1"})

    # Verify notifications were captured
    assert len(received) == 3
    assert received[0].level == NotificationLevel.INFO
    assert received[1].level == NotificationLevel.MILESTONE
    assert received[2].level == NotificationLevel.SUCCESS

    # Send notifications through gateway
    for n in received:
        await gateway.notify(f"{n.title}: {n.body}")

    assert mock_channel.broadcast.call_count == 3


async def test_gateway_routes_messages_to_agent() -> None:
    """Channel message → Gateway → Task handler."""
    events = EventEmitter()
    gateway = ChannelGateway(events)

    tasks_received: list[str] = []

    async def handle_message(msg: ChannelMessage) -> None:
        tasks_received.append(msg.text)

    gateway.set_message_handler(handle_message)

    # Simulate messages from different channels
    msg1 = ChannelMessage(text="build a website", channel=ChannelType.TELEGRAM, sender_id="tg-1")
    msg2 = ChannelMessage(text="deploy to AWS", channel=ChannelType.DISCORD, sender_id="dc-1")

    await gateway.route_message(msg1)
    await gateway.route_message(msg2)

    assert len(tasks_received) == 2
    assert "build a website" in tasks_received
    assert "deploy to AWS" in tasks_received


async def test_notifier_with_error_events() -> None:
    """Error events generate error-level notifications."""
    events = EventEmitter()
    notifier = Notifier(events)

    errors: list[Notification] = []

    async def error_sink(n: Notification) -> None:
        if n.level == NotificationLevel.ERROR:
            errors.append(n)

    notifier.add_sink(error_sink)
    notifier.subscribe()

    await events.emit("task.failed", {"task_id": "t-1", "error": "LLM timeout"})
    await events.emit("ledger.gate_failed", {"gate": 5}, task_id="t-1")

    assert len(errors) == 2
    assert "LLM timeout" in errors[0].body
    assert "Gate" in errors[1].body



"""Tests for CopilotEngine interactive features — plan state, multi-turn, user input, delivery."""

from __future__ import annotations

from pathlib import Path

import pytest

from hauba.engine.copilot_engine import CopilotEngine, PlanState
from hauba.engine.types import EngineConfig, EngineResult, ProviderType

# --- PlanState tests ---


def test_plan_state_save_load(tmp_path: Path) -> None:
    """PlanState can be saved to disk and loaded back."""
    state = PlanState(
        task="Build a REST API",
        plan_text="1. Create FastAPI app\n2. Add endpoints",
        approved=True,
        session_id="test-session-123",
        timestamp=1234567890.0,
        workspace="/tmp/hauba-output",
        files_created=["main.py", "requirements.txt"],
    )

    plan_path = tmp_path / "plan.json"
    state.save(plan_path)

    assert plan_path.exists()

    loaded = PlanState.load(plan_path)
    assert loaded is not None
    assert loaded.task == "Build a REST API"
    assert loaded.approved is True
    assert loaded.session_id == "test-session-123"
    assert loaded.files_created == ["main.py", "requirements.txt"]
    assert loaded.plan_text == "1. Create FastAPI app\n2. Add endpoints"


def test_plan_state_load_nonexistent(tmp_path: Path) -> None:
    """Loading from a nonexistent path returns None."""
    result = PlanState.load(tmp_path / "nonexistent.json")
    assert result is None


def test_plan_state_load_invalid_json(tmp_path: Path) -> None:
    """Loading invalid JSON returns None."""
    plan_path = tmp_path / "bad.json"
    plan_path.write_text("not json!!!", encoding="utf-8")
    result = PlanState.load(plan_path)
    assert result is None


def test_plan_state_defaults() -> None:
    """PlanState has sensible defaults."""
    state = PlanState()
    assert state.task == ""
    assert state.plan_text == ""
    assert state.approved is False
    assert state.session_id == ""
    assert state.files_created == []


# --- Engine interactive handler wiring tests ---


def test_set_user_input_handler() -> None:
    """User input handler can be set and is stored."""
    config = EngineConfig(provider=ProviderType.ANTHROPIC, api_key="test")
    engine = CopilotEngine(config)

    async def handler(q: str, c: list[str], f: bool) -> str:
        return "test"

    engine.set_user_input_handler(handler)
    assert engine._user_input_handler is handler


def test_set_plan_review_handler() -> None:
    """Plan review handler can be set and is stored."""
    config = EngineConfig(provider=ProviderType.ANTHROPIC, api_key="test")
    engine = CopilotEngine(config)

    async def handler(plan: str) -> bool:
        return True

    engine.set_plan_review_handler(handler)
    assert engine._plan_review_handler is handler


def test_set_delivery_handler() -> None:
    """Delivery handler can be set and is stored."""
    config = EngineConfig(provider=ProviderType.ANTHROPIC, api_key="test")
    engine = CopilotEngine(config)

    async def handler(output: str, sid: str) -> None:
        pass

    engine.set_delivery_handler(handler)
    assert engine._delivery_handler is handler


def test_engine_session_initially_none() -> None:
    """Engine session is None before execute()."""
    config = EngineConfig(provider=ProviderType.ANTHROPIC, api_key="test")
    engine = CopilotEngine(config)
    assert engine.session is None


def test_engine_plan_state_initially_none() -> None:
    """Engine plan_state is None before execute()."""
    config = EngineConfig(provider=ProviderType.ANTHROPIC, api_key="test")
    engine = CopilotEngine(config)
    assert engine.plan_state is None


# --- System prompt now includes human escalation instructions ---


def test_system_prompt_includes_human_escalation() -> None:
    """System prompt tells agent to use ask_user for human escalation."""
    config = EngineConfig(provider=ProviderType.ANTHROPIC, api_key="test")
    engine = CopilotEngine(config)
    prompt = engine._build_hauba_system_prompt()
    assert "HUMAN ESCALATION" in prompt
    assert "ask_user" in prompt
    assert "API key" in prompt
    assert "credentials" in prompt


# --- Session config includes on_user_input_request when handler is set ---


def test_session_config_includes_user_input_handler() -> None:
    """Session config includes on_user_input_request when handler is set."""
    config = EngineConfig(provider=ProviderType.ANTHROPIC, api_key="test")
    engine = CopilotEngine(config)

    async def handler(q: str, c: list[str], f: bool) -> str:
        return "answer"

    engine.set_user_input_handler(handler)
    session_config = engine._build_session_config()
    assert "on_user_input_request" in session_config


def test_session_config_no_user_input_handler_when_not_set() -> None:
    """Session config does NOT include on_user_input_request when no handler."""
    config = EngineConfig(provider=ProviderType.ANTHROPIC, api_key="test")
    engine = CopilotEngine(config)
    session_config = engine._build_session_config()
    assert "on_user_input_request" not in session_config


# --- send_message returns error when no session ---


@pytest.mark.asyncio
async def test_send_message_no_session() -> None:
    """send_message() returns error when no session exists."""
    config = EngineConfig(provider=ProviderType.ANTHROPIC, api_key="test")
    engine = CopilotEngine(config)
    result = await engine.send_message("hello")
    assert not result.success
    assert "No active session" in (result.error or "")


# --- EngineResult tests ---


def test_engine_result_ok() -> None:
    """EngineResult.ok() creates a success result."""
    r = EngineResult.ok(output="done", session_id="s123")
    assert r.success is True
    assert r.output == "done"
    assert r.session_id == "s123"


def test_engine_result_fail() -> None:
    """EngineResult.fail() creates a failure result."""
    r = EngineResult.fail("something broke")
    assert r.success is False
    assert r.error == "something broke"



"""Tests for WebSearchTool — DuckDuckGo + Brave Search."""

from __future__ import annotations

from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from hauba.tools.web import WebSearchTool


@pytest.fixture
def search_tool():
    return WebSearchTool()


# --- Validation ---


async def test_search_requires_query(search_tool):
    result = await search_tool.execute()
    assert not result.success
    assert "Query required" in result.error


# --- DuckDuckGo ---


async def test_ddg_search_returns_results(search_tool):
    """Mock DuckDuckGo HTML response and verify parsing."""
    fake_html = """
    <div class="result">
        <a class="result__a" href="https://example.com">Example Title</a>
        <a class="result__snippet">This is a snippet about the topic.</a>
    </div>
    <div class="result">
        <a class="result__a" href="https://other.com">Other Title</a>
        <a class="result__snippet">Another snippet here.</a>
    </div>
    """
    mock_resp = MagicMock()
    mock_resp.text = fake_html
    mock_resp.raise_for_status = MagicMock()

    mock_client = AsyncMock()
    mock_client.__aenter__ = AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = AsyncMock(return_value=False)
    mock_client.post = AsyncMock(return_value=mock_resp)

    with (
        patch("hauba.tools.web.httpx.AsyncClient", return_value=mock_client),
        patch.dict("os.environ", {}, clear=False),
    ):
        # Ensure no BRAVE_API_KEY
        import os

        orig = os.environ.pop("BRAVE_API_KEY", None)
        try:
            result = await search_tool.execute(query="test query")
        finally:
            if orig is not None:
                os.environ["BRAVE_API_KEY"] = orig

    assert result.success
    assert "Example Title" in result.output


async def test_ddg_no_results(search_tool):
    mock_resp = MagicMock()
    mock_resp.text = "<html><body>No results</body></html>"
    mock_resp.raise_for_status = MagicMock()

    mock_client = AsyncMock()
    mock_client.__aenter__ = AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = AsyncMock(return_value=False)
    mock_client.post = AsyncMock(return_value=mock_resp)

    with patch("hauba.tools.web.httpx.AsyncClient", return_value=mock_client):
        import os

        orig = os.environ.pop("BRAVE_API_KEY", None)
        try:
            result = await search_tool.execute(query="noresults12345")
        finally:
            if orig is not None:
                os.environ["BRAVE_API_KEY"] = orig

    assert result.success
    assert "No results" in result.output


# --- Brave Search ---


async def test_brave_search_used_when_key_set(search_tool):
    """Brave Search API is used when BRAVE_API_KEY is set."""
    brave_response = {
        "web": {
            "results": [
                {"title": "Brave Result", "description": "From Brave", "url": "https://brave.com/r"}
            ]
        }
    }
    mock_resp = MagicMock()
    mock_resp.json = MagicMock(return_value=brave_response)
    mock_resp.raise_for_status = MagicMock()

    mock_client = AsyncMock()
    mock_client.__aenter__ = AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = AsyncMock(return_value=False)
    mock_client.get = AsyncMock(return_value=mock_resp)

    with (
        patch("hauba.tools.web.httpx.AsyncClient", return_value=mock_client),
        patch.dict("os.environ", {"BRAVE_API_KEY": "test-key"}),
    ):
        result = await search_tool.execute(query="brave test")

    assert result.success
    assert "Brave Result" in result.output


# --- Error handling ---


async def test_search_network_error(search_tool):
    mock_client = AsyncMock()
    mock_client.__aenter__ = AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = AsyncMock(return_value=False)
    mock_client.post = AsyncMock(side_effect=Exception("Connection failed"))

    with patch("hauba.tools.web.httpx.AsyncClient", return_value=mock_client):
        import os

        orig = os.environ.pop("BRAVE_API_KEY", None)
        try:
            result = await search_tool.execute(query="test")
        finally:
            if orig is not None:
                os.environ["BRAVE_API_KEY"] = orig

    assert not result.success
    assert "Connection failed" in result.error


# --- DDG URL parsing ---


def test_ddg_html_parsing_with_uddg_redirect(search_tool):
    """DuckDuckGo wraps URLs — test extraction of real URL."""
    html = """
    <a class="result__a" href="//duckduckgo.com/l/?uddg=https%3A%2F%2Freal.com&rut=abc">Real Title</a>
    <a class="result__snippet">A real snippet.</a>
    """
    results = search_tool._parse_ddg_html(html, 5)
    assert len(results) == 1
    assert results[0].url == "https://real.com"
    assert results[0].title == "Real Title"


def test_ddg_html_parsing_limits_results(search_tool):
    html = ""
    for i in range(10):
        html += f'<a class="result__a" href="https://ex{i}.com">Title {i}</a>'
        html += f'<a class="result__snippet">Snippet {i}</a>'
    results = search_tool._parse_ddg_html(html, 3)
    assert len(results) == 3
    assert results[2].rank == 3



"""Core constants for Hauba."""

from __future__ import annotations

from pathlib import Path

# Default Hauba home directory
HAUBA_HOME = Path.home() / ".hauba"

# Subdirectories
AGENTS_DIR = HAUBA_HOME / "agents"
MEMORY_DIR = HAUBA_HOME / "memory"
SKILLS_DIR = HAUBA_HOME / "skills"
LOGS_DIR = HAUBA_HOME / "logs"
BACKUPS_DIR = HAUBA_HOME / "backups"

# Config files
SETTINGS_FILE = HAUBA_HOME / "settings.json"
KEYS_FILE = HAUBA_HOME / "keys.json"
DB_FILE = HAUBA_HOME / "hauba.db"
LAST_SESSION_FILE = HAUBA_HOME / "last_session.json"

# Memory subdirs
OWNER_MEMORY_DIR = MEMORY_DIR / "owner"
KNOWLEDGE_DIR = MEMORY_DIR / "knowledge"
CONTEXT_DIR = MEMORY_DIR / "context"

# Defaults
DEFAULT_LLM_PROVIDER = "anthropic"
DEFAULT_LLM_MODEL = "claude-sonnet-4-5-20250929"
DEFAULT_MAX_TOKENS = 4096
DEFAULT_TEMPERATURE = 0.7

# Event types — task lifecycle
EVENT_TASK_CREATED = "task.created"
EVENT_TASK_STARTED = "task.started"
EVENT_TASK_COMPLETED = "task.completed"
EVENT_TASK_FAILED = "task.failed"

# Engine events
EVENT_ENGINE_STARTED = "engine.started"
EVENT_ENGINE_EXECUTING = "engine.executing"
EVENT_ENGINE_COMPLETED = "engine.completed"
EVENT_ENGINE_ERROR = "engine.error"
EVENT_ENGINE_TIMEOUT = "engine.timeout"

# Tool events
EVENT_TOOL_CALLED = "tool.called"
EVENT_TOOL_RESULT = "tool.result"

# TaskLedger events
EVENT_LEDGER_CREATED = "ledger.created"
EVENT_LEDGER_TASK_STARTED = "ledger.task_started"
EVENT_LEDGER_TASK_COMPLETED = "ledger.task_completed"
EVENT_LEDGER_GATE_PASSED = "ledger.gate_passed"
EVENT_LEDGER_GATE_FAILED = "ledger.gate_failed"

# Quality gates
EVENT_QUALITY_CHECK = "quality.check"
EVENT_QUALITY_PASSED = "quality.passed"
EVENT_QUALITY_FAILED = "quality.failed"

# Browser events
EVENT_BROWSER_NAVIGATE = "browser.navigate"
EVENT_BROWSER_CLICK = "browser.click"
EVENT_BROWSER_TYPE = "browser.type"
EVENT_BROWSER_SCREENSHOT = "browser.screenshot"

# Screen control events
EVENT_SCREEN_CAPTURE = "screen.capture"
EVENT_SCREEN_CLICK = "screen.click"
EVENT_SCREEN_TYPE = "screen.type"

# Replay events
EVENT_REPLAY_STARTED = "replay.started"
EVENT_REPLAY_ENTRY = "replay.entry"
EVENT_REPLAY_FINISHED = "replay.finished"

# Web search events
EVENT_SEARCH_QUERY = "search.query"

# Screenshots directory
SCREENSHOTS_DIR = AGENTS_DIR / "screenshots"

# Screen defaults
DEFAULT_SCREEN_DELAY = 0.5
EMERGENCY_STOP_FILE = HAUBA_HOME / "STOP"

# Voice events
EVENT_VOICE_LISTENING = "voice.listening"
EVENT_VOICE_TRANSCRIBED = "voice.transcribed"
EVENT_VOICE_SPEAKING = "voice.speaking"

# Channel events
EVENT_CHANNEL_MESSAGE = "channel.message"
EVENT_CHANNEL_CONNECTED = "channel.connected"
EVENT_CHANNEL_DISCONNECTED = "channel.disconnected"

# Notification events
EVENT_NOTIFICATION_SENT = "notification.sent"

# Compose events
EVENT_COMPOSE_STARTED = "compose.started"
EVENT_COMPOSE_AGENT_CREATED = "compose.agent_created"
EVENT_COMPOSE_AGENT_STARTED = "compose.agent_started"
EVENT_COMPOSE_AGENT_COMPLETED = "compose.agent_completed"
EVENT_COMPOSE_AGENT_FAILED = "compose.agent_failed"
EVENT_COMPOSE_COMPLETED = "compose.completed"
EVENT_COMPOSE_FAILED = "compose.failed"

# Skill events
EVENT_SKILL_INSTALLED = "skill.installed"
EVENT_SKILL_MATCHED = "skill.matched"

# Bundled skills directory (shipped inside the wheel)
BUNDLED_SKILLS_DIR = Path(__file__).resolve().parent.parent / "bundled_skills"

# Plugin system
PLUGINS_DIR = HAUBA_HOME / "plugins"
BUNDLED_PLUGINS_DIR = Path(__file__).resolve().parent.parent / "bundled_plugins"

# Plugin events
EVENT_PLUGIN_LOADED = "plugin.loaded"
EVENT_PLUGIN_UNLOADED = "plugin.unloaded"
EVENT_PLUGIN_ERROR = "plugin.error"

# Email events
EVENT_EMAIL_SENT = "email.sent"
EVENT_EMAIL_FAILED = "email.failed"

# Auto-reply events
EVENT_AUTOREPLY_TRIGGERED = "autoreply.triggered"
EVENT_AUTOREPLY_ENABLED = "autoreply.enabled"
EVENT_AUTOREPLY_DISABLED = "autoreply.disabled"

# Web fetch events
EVENT_WEB_FETCH = "web.fetch"



"""Type definitions for the Hauba Engine."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Literal


class ProviderType(str, Enum):
    """Supported LLM provider types for BYOK."""

    ANTHROPIC = "anthropic"
    OPENAI = "openai"
    AZURE = "azure"
    OLLAMA = "ollama"


@dataclass
class EngineConfig:
    """Configuration for a Hauba Engine session.

    Users bring their own API key — Hauba owner pays nothing.

    Attributes:
        provider: The LLM provider type.
        api_key: User's API key for the provider.
        model: Model identifier (e.g., "claude-sonnet-4.5", "gpt-4").
        base_url: Optional custom API endpoint URL.
        working_directory: Directory where the agent operates.
        skill_directories: Directories to load skills from.
        streaming: Whether to stream responses.
        copilot_cli_path: Path to copilot CLI binary.
    """

    provider: ProviderType = ProviderType.ANTHROPIC
    api_key: str = ""
    model: str = "claude-sonnet-4-5-20250514"
    base_url: str | None = None
    working_directory: str = "."
    skill_directories: list[str] = field(default_factory=list)
    streaming: bool = True
    copilot_cli_path: str | None = None
    session_persist: bool = True
    auto_install_deps: bool = True

    def to_provider_config(self) -> dict[str, Any]:
        """Convert to Copilot SDK ProviderConfig format."""
        config: dict[str, Any] = {}

        if self.provider == ProviderType.ANTHROPIC:
            config["type"] = "anthropic"
            config["base_url"] = self.base_url or "https://api.anthropic.com"
            if self.api_key:
                config["api_key"] = self.api_key
        elif self.provider == ProviderType.OPENAI:
            config["type"] = "openai"
            config["base_url"] = self.base_url or "https://api.openai.com/v1"
            if self.api_key:
                config["api_key"] = self.api_key
        elif self.provider == ProviderType.AZURE:
            config["type"] = "azure"
            if self.base_url:
                config["base_url"] = self.base_url
            if self.api_key:
                config["api_key"] = self.api_key
        elif self.provider == ProviderType.OLLAMA:
            config["type"] = "openai"
            config["base_url"] = self.base_url or "http://localhost:11434/v1"
            # Ollama doesn't need an API key

        return config


@dataclass
class EngineEvent:
    """An event emitted during engine execution.

    Attributes:
        type: Event type string (e.g., "assistant.message", "tool.execution_start").
        data: Event-specific data.
        timestamp: When the event occurred.
    """

    type: str
    data: Any = None
    timestamp: float = 0.0


@dataclass
class EngineResult:
    """Result of an engine execution.

    Attributes:
        success: Whether the task completed successfully.
        output: The final output text.
        events: List of events that occurred during execution.
        error: Error message if failed.
        session_id: The Copilot session ID (for resumption).
    """

    success: bool
    output: str = ""
    events: list[EngineEvent] = field(default_factory=list)
    error: str | None = None
    session_id: str | None = None

    @staticmethod
    def ok(output: str, session_id: str | None = None) -> EngineResult:
        return EngineResult(success=True, output=output, session_id=session_id)

    @staticmethod
    def fail(error: str) -> EngineResult:
        return EngineResult(success=False, error=error)


# Status of a running task
TaskExecutionStatus = Literal["pending", "running", "completed", "failed", "cancelled"]



"""Git tool — git operations via subprocess."""

from __future__ import annotations

from typing import Any

import structlog

from hauba.core.types import ToolResult
from hauba.tools.base import BaseTool
from hauba.tools.bash import BashTool

logger = structlog.get_logger()


class GitTool(BaseTool):
    """Git operations: status, add, commit, push, pull, diff, log."""

    name = "git"
    description = (
        "Execute git operations like status, add, commit, push, pull, diff, log, and init."
    )

    def __init__(self, cwd: str | None = None) -> None:
        self._bash = BashTool(cwd=cwd)

    def _parameters_schema(self) -> dict[str, Any]:
        return {
            "type": "object",
            "properties": {
                "action": {
                    "type": "string",
                    "enum": ["status", "add", "commit", "push", "pull", "diff", "log", "init"],
                    "description": "The git operation to perform.",
                },
                "files": {
                    "type": "string",
                    "description": "Files to add (for 'add' action). Default '.'",
                },
                "message": {
                    "type": "string",
                    "description": "Commit message (for 'commit' action).",
                },
                "args": {
                    "type": "string",
                    "description": "Additional git arguments.",
                },
            },
            "required": ["action"],
        }

    async def execute(self, **kwargs: object) -> ToolResult:
        """Execute a git operation."""
        action = str(kwargs.get("action", ""))
        if not action:
            return ToolResult(tool_name=self.name, success=False, error="No action specified")

        cmd = self._build_command(action, kwargs)
        if cmd is None:
            return ToolResult(tool_name=self.name, success=False, error=f"Unknown action: {action}")

        result = await self._bash.execute(command=cmd)
        return ToolResult(
            tool_name=self.name,
            success=result.success,
            output=result.output,
            error=result.error,
            exit_code=result.exit_code,
        )

    def _build_command(self, action: str, kwargs: dict) -> str | None:
        """Build the git command string."""
        if action == "status":
            return "git status"
        elif action == "add":
            files = str(kwargs.get("files", "."))
            return f"git add {files}"
        elif action == "commit":
            message = str(kwargs.get("message", "Auto-commit by Hauba"))
            safe_msg = message.replace('"', '\\"')
            return f'git commit -m "{safe_msg}"'
        elif action == "push":
            return "git push"
        elif action == "pull":
            return "git pull"
        elif action == "diff":
            return "git diff"
        elif action == "log":
            return "git log --oneline -10"
        elif action == "init":
            return "git init"
        else:
            return None



"""Tests for Notifier — event-driven notification system."""

from __future__ import annotations

from unittest.mock import AsyncMock

import pytest

from hauba.core.events import EventEmitter
from hauba.core.notifier import Notification, NotificationLevel, Notifier


@pytest.fixture
def events() -> EventEmitter:
    return EventEmitter()


@pytest.fixture
def notifier(events: EventEmitter) -> Notifier:
    n = Notifier(events)
    n.subscribe()
    return n


# --- Notification model ---


def test_notification_to_dict() -> None:
    n = Notification(
        level=NotificationLevel.SUCCESS,
        title="Done",
        body="Task complete",
        task_id="t-1",
        source="dir-1",
    )
    d = n.to_dict()
    assert d["level"] == "success"
    assert d["title"] == "Done"
    assert d["body"] == "Task complete"
    assert "timestamp" in d


# --- Send to sinks ---


async def test_send_to_single_sink(notifier: Notifier) -> None:
    sink = AsyncMock()
    notifier.add_sink(sink)

    n = Notification(level=NotificationLevel.INFO, title="Test")
    sent = await notifier.send(n)
    assert sent == 1
    sink.assert_called_once_with(n)


async def test_send_to_multiple_sinks(notifier: Notifier) -> None:
    sink1 = AsyncMock()
    sink2 = AsyncMock()
    notifier.add_sink(sink1)
    notifier.add_sink(sink2)

    n = Notification(level=NotificationLevel.ERROR, title="Failure")
    sent = await notifier.send(n)
    assert sent == 2


async def test_send_handles_sink_error(notifier: Notifier) -> None:
    bad_sink = AsyncMock(side_effect=Exception("sink error"))
    good_sink = AsyncMock()
    notifier.add_sink(bad_sink)
    notifier.add_sink(good_sink)

    n = Notification(level=NotificationLevel.INFO, title="Test")
    sent = await notifier.send(n)
    assert sent == 1  # Only good sink succeeded


# --- Event-driven notifications ---


async def test_task_started_generates_notification(
    events: EventEmitter, notifier: Notifier
) -> None:
    sink = AsyncMock()
    notifier.add_sink(sink)

    await events.emit("task.started", {"task_id": "t-1", "instruction": "build app"})
    assert sink.call_count == 1
    notification = sink.call_args[0][0]
    assert notification.level == NotificationLevel.INFO
    assert notification.title == "Task Started"


async def test_task_completed_generates_notification(
    events: EventEmitter, notifier: Notifier
) -> None:
    sink = AsyncMock()
    notifier.add_sink(sink)

    await events.emit("task.completed", {"task_id": "t-2"})
    notification = sink.call_args[0][0]
    assert notification.level == NotificationLevel.SUCCESS
    assert notification.title == "Task Completed"


async def test_task_failed_generates_notification(events: EventEmitter, notifier: Notifier) -> None:
    sink = AsyncMock()
    notifier.add_sink(sink)

    await events.emit("task.failed", {"task_id": "t-3", "error": "timeout"})
    notification = sink.call_args[0][0]
    assert notification.level == NotificationLevel.ERROR
    assert "timeout" in notification.body


async def test_milestone_completed_notification(events: EventEmitter, notifier: Notifier) -> None:
    sink = AsyncMock()
    notifier.add_sink(sink)

    await events.emit("milestone.completed", {"milestone_id": "m-1"}, task_id="t-1")
    notification = sink.call_args[0][0]
    assert notification.level == NotificationLevel.MILESTONE


async def test_gate_failed_notification(events: EventEmitter, notifier: Notifier) -> None:
    sink = AsyncMock()
    notifier.add_sink(sink)

    await events.emit("ledger.gate_failed", {"gate": 4}, task_id="t-1")
    notification = sink.call_args[0][0]
    assert notification.level == NotificationLevel.ERROR
    assert "Gate" in notification.body


# --- History ---


async def test_notification_history(notifier: Notifier) -> None:
    sink = AsyncMock()
    notifier.add_sink(sink)

    for i in range(5):
        n = Notification(level=NotificationLevel.INFO, title=f"Notif {i}")
        await notifier.send(n)

    history = notifier.get_history()
    assert len(history) == 5


async def test_notification_history_filtered(notifier: Notifier) -> None:
    sink = AsyncMock()
    notifier.add_sink(sink)

    await notifier.send(Notification(level=NotificationLevel.INFO, title="Info"))
    await notifier.send(Notification(level=NotificationLevel.ERROR, title="Error"))
    await notifier.send(Notification(level=NotificationLevel.INFO, title="Info 2"))

    errors = notifier.get_history(level=NotificationLevel.ERROR)
    assert len(errors) == 1
    assert errors[0].title == "Error"


# --- Subscribe idempotency ---


def test_subscribe_idempotent(events: EventEmitter) -> None:
    n = Notifier(events)
    n.subscribe()
    n.subscribe()  # Should not double-subscribe
    # Check handler count for one event
    assert len(events._handlers.get("task.started", [])) == 1



"""Hauba directory structure setup."""

from __future__ import annotations

from pathlib import Path

import structlog

from hauba.core.constants import (
    AGENTS_DIR,
    BACKUPS_DIR,
    CONTEXT_DIR,
    HAUBA_HOME,
    KNOWLEDGE_DIR,
    LOGS_DIR,
    MEMORY_DIR,
    OWNER_MEMORY_DIR,
    SKILLS_DIR,
)

logger = structlog.get_logger()

REQUIRED_DIRS = [
    HAUBA_HOME,
    AGENTS_DIR,
    MEMORY_DIR,
    OWNER_MEMORY_DIR,
    KNOWLEDGE_DIR,
    CONTEXT_DIR,
    SKILLS_DIR,
    LOGS_DIR,
    BACKUPS_DIR,
]


def ensure_hauba_dirs(base: Path | None = None) -> Path:
    """Create the ~/.hauba/ directory structure. Returns the base path."""
    home = base or HAUBA_HOME
    dirs = (
        REQUIRED_DIRS
        if base is None
        else [
            home,
            home / "agents",
            home / "memory",
            home / "memory" / "owner",
            home / "memory" / "knowledge",
            home / "memory" / "context",
            home / "skills",
            home / "logs",
            home / "backups",
        ]
    )
    for d in dirs:
        d.mkdir(parents=True, exist_ok=True)
    logger.info("setup.dirs_created", home=str(home))
    return home



"""Hauba API — HTTP service for the Hauba AI Engineer.

BYOK model: Users bring their own LLM API key. Hauba owner pays nothing.
Deployed on Railway (or any platform).
"""

from hauba.api.server import create_app

__all__ = ["create_app"]



"""Tests for Write-Ahead Log (WAL) — crash-safe state persistence."""

from __future__ import annotations

from pathlib import Path

from hauba.ledger.wal import (
    WAL_OP_ADD_TASK,
    WAL_OP_COMPLETE_TASK,
    WAL_OP_START_TASK,
    WriteAheadLog,
)


def test_append_and_replay(tmp_path: Path) -> None:
    wal = WriteAheadLog(tmp_path / "test.wal")
    wal.append(WAL_OP_ADD_TASK, "task-1", {"desc": "First task"})
    wal.append(WAL_OP_START_TASK, "task-1")
    wal.append(WAL_OP_COMPLETE_TASK, "task-1", {"hash": "abc123"})

    assert wal.entry_count == 3
    assert wal.sequence == 3


def test_persist_to_disk(tmp_path: Path) -> None:
    wal_path = tmp_path / "persist.wal"
    wal = WriteAheadLog(wal_path)
    wal.append(WAL_OP_ADD_TASK, "X", {"desc": "Task X"})
    assert wal_path.exists()

    # Reload from disk
    wal2 = WriteAheadLog(wal_path)
    assert wal2.entry_count == 1
    entries = wal2.replay_all()
    assert entries[0].task_id == "X"


def test_checkpoint_and_replay(tmp_path: Path) -> None:
    wal = WriteAheadLog(tmp_path / "cp.wal")
    wal.append(WAL_OP_ADD_TASK, "A", {})
    wal.append(WAL_OP_COMPLETE_TASK, "A", {})
    wal.checkpoint()
    wal.append(WAL_OP_ADD_TASK, "B", {})

    # Replay from checkpoint should only return B
    entries = wal.replay_from_checkpoint()
    assert len(entries) == 1
    assert entries[0].task_id == "B"


def test_compact(tmp_path: Path) -> None:
    wal = WriteAheadLog(tmp_path / "compact.wal")
    wal.append(WAL_OP_ADD_TASK, "old", {})
    wal.checkpoint()
    wal.append(WAL_OP_ADD_TASK, "new", {})

    wal.compact()
    assert wal.entry_count == 1
    entries = wal.replay_all()
    assert entries[0].task_id == "new"


def test_empty_wal(tmp_path: Path) -> None:
    wal = WriteAheadLog(tmp_path / "empty.wal")
    assert wal.entry_count == 0
    assert wal.replay_all() == []
    assert wal.replay_from_checkpoint() == []



"""Skill CLI — Manage Hauba skills from the command line."""

from __future__ import annotations

import shutil
from pathlib import Path

import typer
from rich.console import Console
from rich.panel import Panel
from rich.table import Table

from hauba.core.constants import BUNDLED_SKILLS_DIR, SKILLS_DIR
from hauba.skills.loader import SkillLoader

skill_app = typer.Typer(
    name="skill",
    help="Manage Hauba skills",
    no_args_is_help=True,
)
console = Console()


@skill_app.command(name="list")
def list_skills() -> None:
    """List all installed and bundled skills."""
    loader = SkillLoader(skill_dirs=[SKILLS_DIR, BUNDLED_SKILLS_DIR])
    skills = loader.load_all()

    if not skills:
        console.print("[dim]No skills found.[/dim]")
        return

    table = Table(title="Available Skills")
    table.add_column("Name", style="cyan")
    table.add_column("Source", style="dim")
    table.add_column("Capabilities", style="green")

    for name in sorted(skills.keys()):
        skill = skills[name]
        source = "bundled" if BUNDLED_SKILLS_DIR in skill.file_path.parents else "installed"
        caps = "; ".join(skill.capabilities[:2])
        if len(skill.capabilities) > 2:
            caps += f" (+{len(skill.capabilities) - 2} more)"
        table.add_row(name, source, caps)

    console.print(table)
    console.print(f"\n[dim]{len(skills)} skill(s) available[/dim]")


@skill_app.command()
def install(path: str = typer.Argument(..., help="Path to a .md skill file")) -> None:
    """Install a skill from a .md file."""
    source = Path(path)
    if not source.exists():
        console.print(f"[red]File not found: {path}[/red]")
        raise typer.Exit(1)
    if source.suffix != ".md":
        console.print("[red]Skill files must be .md files[/red]")
        raise typer.Exit(1)

    SKILLS_DIR.mkdir(parents=True, exist_ok=True)
    dest = SKILLS_DIR / source.name
    shutil.copy2(source, dest)
    console.print(f"[green]Installed skill: {source.stem} -> {dest}[/green]")


@skill_app.command()
def create(name: str = typer.Argument(..., help="Name for the new skill")) -> None:
    """Scaffold a new skill .md template."""
    slug = name.lower().replace(" ", "-")
    filename = f"{slug}.md"
    dest = Path.cwd() / filename

    if dest.exists():
        console.print(f"[red]File already exists: {dest}[/red]")
        raise typer.Exit(1)

    template = f"""# Skill: {slug}

## Capabilities
- Describe what this skill enables

## When To Use
- Describe trigger conditions for this skill

## Approach

### Phase 1: Understand
- How to analyze the problem space

### Phase 2: Plan
- How to decompose into executable steps

### Phase 3: Execute
- Step-by-step execution with decision points

### Phase 4: Verify
- How to validate the output

## Constraints
- Safety rails and anti-patterns to avoid

## Scale Considerations
- What changes at large scale

## Error Recovery
- Common failure modes and how to recover
"""
    dest.write_text(template, encoding="utf-8")
    console.print(f"[green]Created skill template: {dest}[/green]")
    console.print(f"[dim]Edit the file, then install with: hauba skill install {filename}[/dim]")


@skill_app.command()
def show(name: str = typer.Argument(..., help="Skill name to display")) -> None:
    """Display a skill's content."""
    loader = SkillLoader(skill_dirs=[SKILLS_DIR, BUNDLED_SKILLS_DIR])

    try:
        skill = loader.get(name)
    except Exception:
        console.print(f"[red]Skill not found: {name}[/red]")
        raise typer.Exit(1)

    console.print(
        Panel(
            skill.raw_content,
            title=f"Skill: {skill.name}",
            border_style="cyan",
        )
    )



"""Tests for WhatsApp V4.0 commands."""

from __future__ import annotations

from unittest.mock import AsyncMock

import pytest

from hauba.channels.whatsapp_webhook import WhatsAppBot
from hauba.daemon.queue import TaskQueue


@pytest.fixture
def bot() -> WhatsAppBot:
    b = WhatsAppBot()
    b._send_reply = AsyncMock()  # type: ignore[assignment]
    return b


@pytest.fixture
def bot_with_queue(bot: WhatsAppBot) -> WhatsAppBot:
    queue = TaskQueue()
    bot.set_task_queue(queue)
    return bot


class TestParseCommand:
    """Test WhatsAppBot._parse_command()."""

    def test_command_with_args(self) -> None:
        cmd, args = WhatsAppBot._parse_command("/cancel abc123")
        assert cmd == "/cancel"
        assert args == "abc123"

    def test_command_no_args(self) -> None:
        cmd, args = WhatsAppBot._parse_command("/tasks")
        assert cmd == "/tasks"
        assert args == ""

    def test_not_a_command(self) -> None:
        cmd, args = WhatsAppBot._parse_command("build me a dashboard")
        assert cmd == ""
        assert args == "build me a dashboard"

    def test_command_case_insensitive(self) -> None:
        cmd, _args = WhatsAppBot._parse_command("/HELP")
        assert cmd == "/help"


class TestHandleTasks:
    """Test /tasks command."""

    @pytest.mark.asyncio
    async def test_no_queue(self, bot: WhatsAppBot) -> None:
        await bot._handle_tasks("whatsapp:+1234")
        bot._send_reply.assert_called_once()
        assert "No task queue" in bot._send_reply.call_args[0][1]

    @pytest.mark.asyncio
    async def test_no_tasks(self, bot_with_queue: WhatsAppBot) -> None:
        await bot_with_queue._handle_tasks("whatsapp:+1234")
        assert "No tasks" in bot_with_queue._send_reply.call_args[0][1]

    @pytest.mark.asyncio
    async def test_with_tasks(self, bot_with_queue: WhatsAppBot) -> None:
        bot_with_queue._task_queue.submit(
            "whatsapp:+1234",
            "build something",
            channel="whatsapp",
            channel_address="whatsapp:+1234",
        )
        await bot_with_queue._handle_tasks("whatsapp:+1234")
        reply = bot_with_queue._send_reply.call_args[0][1]
        assert "build something" in reply


class TestHandleCancel:
    """Test /cancel command."""

    @pytest.mark.asyncio
    async def test_cancel_no_args(self, bot_with_queue: WhatsAppBot) -> None:
        await bot_with_queue._handle_cancel("whatsapp:+1234", "")
        assert "Usage" in bot_with_queue._send_reply.call_args[0][1]

    @pytest.mark.asyncio
    async def test_cancel_existing_task(self, bot_with_queue: WhatsAppBot) -> None:
        task = bot_with_queue._task_queue.submit("whatsapp:+1234", "build app")
        prefix = task.task_id[:8]
        await bot_with_queue._handle_cancel("whatsapp:+1234", prefix)
        assert "cancelled" in bot_with_queue._send_reply.call_args[0][1].lower()

    @pytest.mark.asyncio
    async def test_cancel_not_found(self, bot_with_queue: WhatsAppBot) -> None:
        await bot_with_queue._handle_cancel("whatsapp:+1234", "nonexist")
        assert "No task found" in bot_with_queue._send_reply.call_args[0][1]


class TestHandleRetry:
    """Test /retry command."""

    @pytest.mark.asyncio
    async def test_retry_no_args(self, bot_with_queue: WhatsAppBot) -> None:
        await bot_with_queue._handle_retry("whatsapp:+1234", "")
        assert "Usage" in bot_with_queue._send_reply.call_args[0][1]

    @pytest.mark.asyncio
    async def test_retry_failed_task(self, bot_with_queue: WhatsAppBot) -> None:
        task = bot_with_queue._task_queue.submit("whatsapp:+1234", "build dashboard")
        task.status = "failed"
        prefix = task.task_id[:8]
        await bot_with_queue._handle_retry("whatsapp:+1234", prefix)
        assert "retried" in bot_with_queue._send_reply.call_args[0][1].lower()


class TestHandleWeb:
    """Test /web command."""

    @pytest.mark.asyncio
    async def test_no_url(self, bot: WhatsAppBot) -> None:
        await bot._handle_web("whatsapp:+1234", "")
        assert "Usage" in bot._send_reply.call_args[0][1]


class TestHandleEmail:
    """Test /email command."""

    @pytest.mark.asyncio
    async def test_no_email_service(self, bot: WhatsAppBot) -> None:
        await bot._handle_email("whatsapp:+1234", "user@test.com subject")
        assert "not configured" in bot._send_reply.call_args[0][1].lower()

    @pytest.mark.asyncio
    async def test_no_args(self, bot: WhatsAppBot) -> None:
        bot._email_service = AsyncMock()
        bot._email_service.is_configured = True
        await bot._handle_email("whatsapp:+1234", "")
        assert "Usage" in bot._send_reply.call_args[0][1]


class TestHandleReply:
    """Test /reply command."""

    @pytest.mark.asyncio
    async def test_no_assistant(self, bot: WhatsAppBot) -> None:
        await bot._handle_reply_cmd("whatsapp:+1234", "I'm away")
        assert "not available" in bot._send_reply.call_args[0][1].lower()

    @pytest.mark.asyncio
    async def test_set_reply(self, bot: WhatsAppBot) -> None:
        from hauba.services.reply_assistant import ReplyAssistant

        bot._reply_assistant = ReplyAssistant()
        await bot._handle_reply_cmd("whatsapp:+1234", "Out of office")
        assert "set" in bot._send_reply.call_args[0][1].lower()

    @pytest.mark.asyncio
    async def test_disable_reply(self, bot: WhatsAppBot) -> None:
        from hauba.services.reply_assistant import ReplyAssistant

        bot._reply_assistant = ReplyAssistant()
        await bot._handle_reply_cmd("whatsapp:+1234", "off")
        assert "disabled" in bot._send_reply.call_args[0][1].lower()


class TestHandleUsage:
    """Test /usage command."""

    @pytest.mark.asyncio
    async def test_no_queue(self, bot: WhatsAppBot) -> None:
        await bot._handle_usage("whatsapp:+1234")
        assert "no usage data" in bot._send_reply.call_args[0][1].lower()

    @pytest.mark.asyncio
    async def test_with_tasks(self, bot_with_queue: WhatsAppBot) -> None:
        bot_with_queue._task_queue.submit("whatsapp:+1234", "task 1")
        await bot_with_queue._handle_usage("whatsapp:+1234")
        reply = bot_with_queue._send_reply.call_args[0][1]
        assert "Total tasks" in reply


class TestHandlePlugins:
    """Test /plugins command."""

    @pytest.mark.asyncio
    async def test_no_registry(self, bot: WhatsAppBot) -> None:
        await bot._handle_plugins("whatsapp:+1234")
        assert "No plugins" in bot._send_reply.call_args[0][1]


class TestHandleFeedback:
    """Test /feedback command."""

    @pytest.mark.asyncio
    async def test_no_message(self, bot: WhatsAppBot) -> None:
        await bot._handle_feedback("whatsapp:+1234", "")
        assert "Usage" in bot._send_reply.call_args[0][1]

    @pytest.mark.asyncio
    async def test_feedback_received(self, bot: WhatsAppBot) -> None:
        await bot._handle_feedback("whatsapp:+1234", "Great product!")
        assert "Thank you" in bot._send_reply.call_args[0][1]



"""Rich-based terminal UI for Hauba — Claude Code style output."""

from __future__ import annotations

import sys

import structlog
from rich.console import Console
from rich.panel import Panel

from hauba.core.constants import (
    EVENT_ENGINE_EXECUTING,
    EVENT_ENGINE_STARTED,
    EVENT_TASK_COMPLETED,
    EVENT_TASK_FAILED,
    EVENT_TOOL_CALLED,
    EVENT_TOOL_RESULT,
)
from hauba.core.events import EventEmitter
from hauba.core.types import Event, Result

logger = structlog.get_logger()

# Use ASCII-safe symbols on Windows legacy console (cp1252 etc.)
_USE_ASCII = sys.platform == "win32" and not sys.stdout.encoding.lower().startswith("utf")

# Symbol map: Unicode -> ASCII fallback
_SYM_BULLET = "*" if _USE_ASCII else "\u25cf"  # ●
_SYM_CHECK = "+" if _USE_ASCII else "\u2713"  # ✓
_SYM_CROSS = "x" if _USE_ASCII else "\u2717"  # ✗
_SYM_THINK = "~" if _USE_ASCII else "\u25d0"  # ◐
_SYM_PLAY = ">" if _USE_ASCII else "\u25b6"  # ▶
_SYM_REVIEW = "~" if _USE_ASCII else "\u25d1"  # ◑
_SYM_BOLT = "!" if _USE_ASCII else "\u26a1"  # ⚡
_SYM_FILE = "#" if _USE_ASCII else "\U0001f4c4"  # 📄
_SYM_GIT = "@" if _USE_ASCII else "\U0001f500"  # 🔀
_SYM_EDIT = "~" if _USE_ASCII else "\u270e"  # ✎


class TerminalUI:
    """Rich terminal UI for showing agent progress — Claude Code style."""

    def __init__(self, console: Console, events: EventEmitter) -> None:
        self._console = console
        self._events = events
        self._tool_count = 0
        self._setup_handlers()

    def _setup_handlers(self) -> None:
        """Subscribe to events for live display."""
        self._events.on(EVENT_ENGINE_STARTED, self._on_thinking)
        self._events.on(EVENT_ENGINE_EXECUTING, self._on_executing)
        self._events.on(EVENT_TOOL_CALLED, self._on_tool_called)
        self._events.on(EVENT_TOOL_RESULT, self._on_tool_result)
        self._events.on(EVENT_TASK_COMPLETED, self._on_completed)
        self._events.on(EVENT_TASK_FAILED, self._on_failed)

    async def show_task_start(self, task: str) -> None:
        """Display task start banner."""
        self._console.print()
        self._console.print(
            Panel(
                f"[bold]{task}[/bold]",
                title=f"[cyan]{_SYM_BULLET} Hauba[/cyan]",
                border_style="cyan",
                padding=(1, 2),
            )
        )
        self._console.print()
        self._tool_count = 0

    async def show_task_result(self, result: Result) -> None:
        """Display final task result."""
        self._console.print()
        if result.success:
            # Show the result value, truncated for readability
            value = str(result.value or "Task completed successfully.")
            # Truncate very long outputs
            if len(value) > 2000:
                value = value[:2000] + "\n\n... (output truncated)"
            self._console.print(
                Panel(
                    value,
                    title=f"[green]{_SYM_CHECK} Task Complete ({self._tool_count} tool calls)[/green]",
                    border_style="green",
                    padding=(1, 2),
                )
            )
        else:
            self._console.print(
                Panel(
                    str(result.error or "Task failed."),
                    title=f"[red]{_SYM_CROSS} Task Failed[/red]",
                    border_style="red",
                    padding=(1, 2),
                )
            )
        self._console.print()

    # --- Event Handlers ---

    async def _on_thinking(self, event: Event) -> None:
        self._console.print(f"[dim]  {_SYM_THINK} Thinking...[/dim]")

    async def _on_executing(self, event: Event) -> None:
        steps = event.data.get("steps", 0)
        self._console.print(f"[cyan]  {_SYM_PLAY} Executing ({steps} steps planned)[/cyan]")

    async def _on_tool_called(self, event: Event) -> None:
        tool = event.data.get("tool", "?")
        args = event.data.get("args", {})
        self._tool_count += 1

        if tool == "bash":
            cmd = args.get("command", "")
            self._console.print(
                f"  [bold yellow]{_SYM_BOLT} bash[/bold yellow] [dim]{cmd[:120]}[/dim]"
            )
        elif tool == "files":
            action = args.get("action", "")
            path = args.get("path", "")
            if action == "write":
                content = args.get("content", "")
                lines = content.count("\n") + 1 if content else 0
                self._console.print(
                    f"  [bold blue]{_SYM_FILE} write[/bold blue] [dim]{path} ({lines} lines)[/dim]"
                )
            elif action == "edit":
                self._console.print(f"  [bold blue]{_SYM_EDIT} edit[/bold blue] [dim]{path}[/dim]")
            elif action == "read":
                self._console.print(f"  [bold blue]{_SYM_FILE} read[/bold blue] [dim]{path}[/dim]")
            else:
                self._console.print(
                    f"  [bold blue]{_SYM_FILE} {action}[/bold blue] [dim]{path}[/dim]"
                )
        elif tool == "git":
            action = args.get("action", "")
            self._console.print(f"  [bold magenta]{_SYM_GIT} git {action}[/bold magenta]")
        else:
            self._console.print(f"  [bold]{tool}[/bold]")

    async def _on_tool_result(self, event: Event) -> None:
        tool = event.data.get("tool", "?")
        success = event.data.get("success", False)
        preview = event.data.get("output_preview", "")

        if success:
            if preview.strip():
                lines = preview.strip().split("\n")
                for line in lines[:3]:
                    self._console.print(f"  [dim]  {line[:120]}[/dim]")
                if len(lines) > 3:
                    self._console.print(f"  [dim]  ... ({len(lines) - 3} more lines)[/dim]")
        else:
            self._console.print(f"  [red]  {_SYM_CROSS} {tool} failed[/red]")

    async def _on_completed(self, event: Event) -> None:
        pass  # Handled by show_task_result

    async def _on_failed(self, event: Event) -> None:
        error = event.data.get("error", "Unknown error")
        self._console.print(f"\n  [red]Error: {error[:200]}[/red]")



"""Tests for TelegramChannel — python-telegram-bot integration."""

from __future__ import annotations

from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from hauba.core.events import EventEmitter

# --- Graceful degradation ---


def test_telegram_unavailable_raises() -> None:
    with patch("hauba.channels.telegram.TELEGRAM_AVAILABLE", False):
        from hauba.channels.telegram import TelegramChannel, TelegramChannelError

        events = EventEmitter()
        with pytest.raises(TelegramChannelError, match="python-telegram-bot not installed"):
            TelegramChannel(token="fake", events=events)


# --- Construction with deps available ---


def test_telegram_channel_creates_with_mock_deps() -> None:
    """When telegram deps are mocked as available, channel constructs."""
    mock_app_builder = MagicMock()
    mock_app = MagicMock()
    mock_app_builder.token.return_value.build.return_value = mock_app

    with (
        patch("hauba.channels.telegram.TELEGRAM_AVAILABLE", True),
        patch("hauba.channels.telegram.Application", create=True),
        patch("hauba.channels.telegram.CommandHandler", create=True),
        patch("hauba.channels.telegram.MessageHandler", create=True),
        patch("hauba.channels.telegram.filters", create=True),
    ):
        events = EventEmitter()
        from hauba.channels.telegram import TelegramChannel

        channel = TelegramChannel(token="test-token", events=events)
        assert channel.token == "test-token"


# --- Event forwarding ---


async def test_telegram_on_task_event_broadcasts() -> None:
    """Task events should be broadcast to all known chats."""
    from hauba.channels.telegram import TelegramChannel
    from hauba.core.types import Event

    with (
        patch("hauba.channels.telegram.TELEGRAM_AVAILABLE", True),
        patch("hauba.channels.telegram.Application", create=True),
        patch("hauba.channels.telegram.CommandHandler", create=True),
        patch("hauba.channels.telegram.MessageHandler", create=True),
        patch("hauba.channels.telegram.filters", create=True),
    ):
        events = EventEmitter()
        channel = TelegramChannel(token="t", events=events)
        channel.broadcast = AsyncMock()
        channel._chat_ids = {123, 456}

        event = Event(topic="task.completed", data={"task_id": "t-1"})
        await channel._on_task_event(event)
        channel.broadcast.assert_called_once()
        assert "t-1" in channel.broadcast.call_args[0][0]



"""Email service — send emails on owner's behalf via SMTP."""

from __future__ import annotations

import os
import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

import structlog

logger = structlog.get_logger()


class EmailService:
    """Send emails via SMTP on owner's behalf.

    Configuration via environment variables:
        HAUBA_SMTP_HOST  — SMTP server hostname (e.g., smtp.gmail.com)
        HAUBA_SMTP_PORT  — SMTP port (default: 587)
        HAUBA_SMTP_USER  — SMTP username / email
        HAUBA_SMTP_PASS  — SMTP password / app password
        HAUBA_EMAIL_FROM — From address (defaults to SMTP_USER)

    For Gmail: use an App Password (not your real password).
    For Outlook: use smtp-mail.outlook.com:587.

    Usage:
        service = EmailService()
        if service.configure():
            await service.send("user@example.com", "Hello", "Body text")
    """

    def __init__(self) -> None:
        self._host: str = ""
        self._port: int = 587
        self._user: str = ""
        self._password: str = ""
        self._from_addr: str = ""
        self._configured: bool = False

    def configure(self) -> bool:
        """Load SMTP configuration from environment variables.

        Returns True if all required vars are present.
        """
        self._host = os.environ.get("HAUBA_SMTP_HOST", "")
        self._port = int(os.environ.get("HAUBA_SMTP_PORT", "587"))
        self._user = os.environ.get("HAUBA_SMTP_USER", "")
        self._password = os.environ.get("HAUBA_SMTP_PASS", "")
        self._from_addr = os.environ.get("HAUBA_EMAIL_FROM", self._user)

        if not self._host or not self._user or not self._password:
            logger.info(
                "email.not_configured",
                has_host=bool(self._host),
                has_user=bool(self._user),
                has_pass=bool(self._password),
            )
            self._configured = False
            return False

        self._configured = True
        logger.info("email.configured", host=self._host, user=self._user)
        return True

    @property
    def is_configured(self) -> bool:
        """Whether the email service is configured."""
        return self._configured

    async def send(
        self,
        to: str,
        subject: str,
        body: str,
        *,
        cc: str = "",
        bcc: str = "",
        reply_to: str = "",
    ) -> bool:
        """Send a plain text email.

        Args:
            to: Recipient email address.
            subject: Email subject line.
            body: Plain text email body.
            cc: CC address (optional).
            bcc: BCC address (optional).
            reply_to: Reply-to address (optional).

        Returns True if sent successfully.
        """
        if not self._configured:
            logger.warning("email.not_configured")
            return False

        msg = MIMEMultipart()
        msg["From"] = self._from_addr
        msg["To"] = to
        msg["Subject"] = subject
        if cc:
            msg["Cc"] = cc
        if reply_to:
            msg["Reply-To"] = reply_to

        msg.attach(MIMEText(body, "plain", "utf-8"))

        recipients = [to]
        if cc:
            recipients.append(cc)
        if bcc:
            recipients.append(bcc)

        try:
            # Use sync smtplib (stdlib, always available)
            # aiosmtplib is optional enhancement
            return self._send_sync(msg, recipients)
        except Exception as exc:
            logger.error("email.send_failed", to=to, error=str(exc))
            return False

    async def send_html(
        self,
        to: str,
        subject: str,
        html_body: str,
    ) -> bool:
        """Send an HTML email.

        Args:
            to: Recipient email address.
            subject: Email subject line.
            html_body: HTML email body.

        Returns True if sent successfully.
        """
        if not self._configured:
            return False

        msg = MIMEMultipart("alternative")
        msg["From"] = self._from_addr
        msg["To"] = to
        msg["Subject"] = subject
        msg.attach(MIMEText(html_body, "html", "utf-8"))

        try:
            return self._send_sync(msg, [to])
        except Exception as exc:
            logger.error("email.send_html_failed", to=to, error=str(exc))
            return False

    def _send_sync(self, msg: MIMEMultipart, recipients: list[str]) -> bool:
        """Send email via stdlib smtplib (synchronous, wrapped in async)."""
        try:
            with smtplib.SMTP(self._host, self._port, timeout=30) as server:
                server.ehlo()
                server.starttls()
                server.ehlo()
                server.login(self._user, self._password)
                server.sendmail(self._from_addr, recipients, msg.as_string())

            logger.info("email.sent", to=recipients[0], subject=msg["Subject"])
            return True
        except smtplib.SMTPAuthenticationError:
            logger.error("email.auth_failed", user=self._user)
            return False
        except smtplib.SMTPException as exc:
            logger.error("email.smtp_error", error=str(exc))
            return False



"""Tests for ReplayRecorder — JSON Lines event recording."""

from __future__ import annotations

import json
from pathlib import Path

import pytest

from hauba.core.events import EventEmitter
from hauba.core.replay import ReplayRecorder
from hauba.core.types import Event, ReplayEntry


@pytest.fixture
def recorder(tmp_path: Path) -> ReplayRecorder:
    return ReplayRecorder(task_id="test-task", output_dir=tmp_path)


# --- Initialization ---


def test_recorder_creates_replay_file(tmp_path: Path) -> None:
    recorder = ReplayRecorder(task_id="init-test", output_dir=tmp_path)
    assert recorder.path.exists()
    assert recorder.path.name == ".hauba-replay"
    recorder.close()


def test_recorder_entry_count_starts_zero(recorder: ReplayRecorder) -> None:
    assert recorder.entry_count == 0
    recorder.close()


# --- Recording events ---


async def test_handle_event_writes_json_line(recorder: ReplayRecorder) -> None:
    event = Event(topic="task.started", data={"key": "value"}, source="agent-1", task_id="t1")
    await recorder.handle_event(event)
    recorder.close()

    lines = recorder.path.read_text(encoding="utf-8").strip().split("\n")
    assert len(lines) == 1
    parsed = json.loads(lines[0])
    assert parsed["topic"] == "task.started"
    assert parsed["data"]["key"] == "value"
    assert parsed["source"] == "agent-1"


async def test_handle_event_increments_count(recorder: ReplayRecorder) -> None:
    for i in range(5):
        event = Event(topic=f"event.{i}", data={"i": i})
        await recorder.handle_event(event)
    assert recorder.entry_count == 5
    recorder.close()


async def test_multiple_events_written_as_jsonl(recorder: ReplayRecorder) -> None:
    for i in range(3):
        event = Event(topic=f"tool.call.{i}", data={"step": i})
        await recorder.handle_event(event)
    recorder.close()

    lines = recorder.path.read_text(encoding="utf-8").strip().split("\n")
    assert len(lines) == 3
    for i, line in enumerate(lines):
        parsed = json.loads(line)
        assert parsed["topic"] == f"tool.call.{i}"


# --- Subscribe to EventEmitter ---


async def test_subscribe_captures_all_events(tmp_path: Path) -> None:
    events = EventEmitter()
    recorder = ReplayRecorder(task_id="sub-test", output_dir=tmp_path)
    recorder.subscribe(events)

    await events.emit("task.started", {"msg": "hello"})
    await events.emit("agent.thinking", {"phase": "deliberating"})
    await events.emit("tool.called", {"tool": "bash"})

    assert recorder.entry_count == 3
    recorder.close()

    entries = ReplayRecorder.load(recorder.path)
    assert len(entries) == 3
    assert entries[0].topic == "task.started"
    assert entries[1].topic == "agent.thinking"
    assert entries[2].topic == "tool.called"


# --- Load ---


async def test_load_returns_replay_entries(recorder: ReplayRecorder) -> None:
    event = Event(topic="ledger.gate_passed", data={"gate": 4}, source="dir-1", task_id="t-1")
    await recorder.handle_event(event)
    recorder.close()

    entries = ReplayRecorder.load(recorder.path)
    assert len(entries) == 1
    assert isinstance(entries[0], ReplayEntry)
    assert entries[0].topic == "ledger.gate_passed"
    assert entries[0].data["gate"] == 4


def test_load_empty_file(tmp_path: Path) -> None:
    replay_file = tmp_path / ".hauba-replay"
    replay_file.write_text("", encoding="utf-8")
    entries = ReplayRecorder.load(replay_file)
    assert entries == []


def test_load_skips_blank_lines(tmp_path: Path) -> None:
    replay_file = tmp_path / ".hauba-replay"
    entry = ReplayEntry(topic="a.b", data={"x": 1})
    replay_file.write_text(
        entry.model_dump_json() + "\n\n" + entry.model_dump_json() + "\n",
        encoding="utf-8",
    )
    entries = ReplayRecorder.load(replay_file)
    assert len(entries) == 2


# --- Close ---


def test_close_is_idempotent(recorder: ReplayRecorder) -> None:
    recorder.close()
    recorder.close()  # Should not raise



"""Base plugin interface for Hauba."""

from __future__ import annotations

from abc import ABC, abstractmethod


class BasePlugin(ABC):
    """Abstract base class for all Hauba plugins.

    Plugins can hook into the Hauba lifecycle to extend functionality.
    Third-party services implement this interface to integrate with Hauba.

    Example:
        class MyPlugin(BasePlugin):
            name = "my-plugin"
            description = "Does something useful"

            async def on_message(self, channel, sender, text):
                if "hello" in text.lower():
                    return "Hi there from my plugin!"
                return None
    """

    name: str = ""
    description: str = ""
    version: str = "0.1.0"

    @abstractmethod
    async def on_load(self) -> None:
        """Called when the plugin is loaded. Initialize resources here."""

    @abstractmethod
    async def on_unload(self) -> None:
        """Called when the plugin is unloaded. Clean up resources here."""

    async def on_message(self, channel: str, sender: str, text: str) -> str | None:
        """Called when a message is received on any channel.

        Return a string to send as a reply, or None to pass through.
        """
        return None

    async def on_task_complete(self, task_id: str, output: str) -> None:
        """Called when a task completes execution."""
        return

    async def on_task_queued(self, task_id: str, instruction: str) -> None:
        """Called when a new task is queued."""
        return

    async def on_startup(self) -> None:
        """Called when the Hauba system starts."""
        return

    async def on_shutdown(self) -> None:
        """Called when the Hauba system shuts down."""
        return



"""Tests for WhatsApp bot queue integration."""

from __future__ import annotations

from unittest.mock import AsyncMock

import pytest

from hauba.channels.whatsapp_webhook import WhatsAppBot
from hauba.daemon.queue import TaskQueue


class TestIsBuildTask:
    """Test build task detection."""

    def test_short_messages_are_chat(self) -> None:
        assert WhatsAppBot._is_build_task("hi") is False
        assert WhatsAppBot._is_build_task("hello") is False
        assert WhatsAppBot._is_build_task("thanks") is False

    def test_build_keywords_detected(self) -> None:
        assert WhatsAppBot._is_build_task("build me a todo app") is True
        assert WhatsAppBot._is_build_task("create a REST API with auth") is True
        assert WhatsAppBot._is_build_task("make a SaaS dashboard") is True
        assert WhatsAppBot._is_build_task("deploy this to Railway") is True

    def test_code_keywords_detected(self) -> None:
        assert WhatsAppBot._is_build_task("write a Python script to parse CSV") is True
        assert WhatsAppBot._is_build_task("generate a landing page with HTML") is True
        assert WhatsAppBot._is_build_task("implement user authentication") is True

    def test_general_chat_not_detected(self) -> None:
        assert WhatsAppBot._is_build_task("what can you do?") is False
        assert WhatsAppBot._is_build_task("how are you today") is False
        assert WhatsAppBot._is_build_task("tell me about hauba") is False


class TestSetTaskQueue:
    """Test task queue wiring."""

    def test_set_task_queue(self) -> None:
        bot = WhatsAppBot()
        queue = TaskQueue()
        bot.set_task_queue(queue)
        assert bot._task_queue is queue

    def test_no_queue_by_default(self) -> None:
        bot = WhatsAppBot()
        assert bot._task_queue is None


class TestQueueBuildTask:
    """Test queuing build tasks."""

    @pytest.mark.asyncio
    async def test_queue_build_task_with_queue(self) -> None:
        bot = WhatsAppBot()
        queue = TaskQueue()
        bot.set_task_queue(queue)
        bot._send_reply = AsyncMock()  # type: ignore[assignment]

        await bot._queue_build_task("whatsapp:+1234", "build me a todo app")

        # Task should be in the queue
        tasks = queue.poll("whatsapp:+1234")
        assert len(tasks) == 1
        assert tasks[0].instruction == "build me a todo app"
        assert tasks[0].channel == "whatsapp"
        assert tasks[0].channel_address == "whatsapp:+1234"

        # Should have sent a confirmation
        bot._send_reply.assert_called_once()
        call_args = bot._send_reply.call_args
        assert "Task Queued" in call_args[0][1]

    @pytest.mark.asyncio
    async def test_queue_build_task_without_queue_falls_back(self) -> None:
        """Without a queue, falls back to direct execution."""
        bot = WhatsAppBot()
        bot._send_reply = AsyncMock()  # type: ignore[assignment]
        bot._process_with_progress = AsyncMock(return_value="Done!")  # type: ignore[assignment]

        await bot._queue_build_task("whatsapp:+1234", "build me an app")

        # Should have used process_with_progress as fallback
        bot._process_with_progress.assert_called_once()


class TestSendStatus:
    """Test status command."""

    @pytest.mark.asyncio
    async def test_status_no_queue(self) -> None:
        bot = WhatsAppBot()
        bot._send_reply = AsyncMock()  # type: ignore[assignment]

        await bot._send_status("whatsapp:+1234")
        bot._send_reply.assert_called_once()
        assert "No task queue" in bot._send_reply.call_args[0][1]

    @pytest.mark.asyncio
    async def test_status_no_tasks(self) -> None:
        bot = WhatsAppBot()
        queue = TaskQueue()
        bot.set_task_queue(queue)
        bot._send_reply = AsyncMock()  # type: ignore[assignment]

        await bot._send_status("whatsapp:+1234")
        bot._send_reply.assert_called_once()
        assert "No tasks found" in bot._send_reply.call_args[0][1]

    @pytest.mark.asyncio
    async def test_status_with_tasks(self) -> None:
        bot = WhatsAppBot()
        queue = TaskQueue()
        bot.set_task_queue(queue)
        bot._send_reply = AsyncMock()  # type: ignore[assignment]

        queue.submit("whatsapp:+1234", "build a todo app", channel="whatsapp")

        await bot._send_status("whatsapp:+1234")
        bot._send_reply.assert_called_once()
        msg = bot._send_reply.call_args[0][1]
        assert "Your Tasks" in msg
        assert "todo app" in msg


class TestStatusCommand:
    """Test the /status command in handle_message."""

    @pytest.mark.asyncio
    async def test_status_command(self) -> None:
        bot = WhatsAppBot()
        queue = TaskQueue()
        bot.set_task_queue(queue)
        bot._send_reply = AsyncMock()  # type: ignore[assignment]

        await bot.handle_message("/status", "whatsapp:+1234")
        bot._send_reply.assert_called_once()

    @pytest.mark.asyncio
    async def test_clear_clears_queue(self) -> None:
        bot = WhatsAppBot()
        queue = TaskQueue()
        bot.set_task_queue(queue)
        bot._send_reply = AsyncMock()  # type: ignore[assignment]

        queue.submit("whatsapp:+1234", "build app", channel="whatsapp")
        assert queue.size == 1

        await bot.handle_message("/clear", "whatsapp:+1234")
        assert queue.size == 0



"""File tool — read, write, edit files and directories."""

from __future__ import annotations

import os
import tempfile
from pathlib import Path
from typing import Any

import structlog

from hauba.core.types import ToolResult
from hauba.tools.base import BaseTool

logger = structlog.get_logger()


class FileTool(BaseTool):
    """Read, write, edit, and manage files."""

    name = "files"
    description = "Read, write, edit files and create directories. Use action='write' to create new files, action='read' to read existing files, action='edit' for search-and-replace edits, action='mkdir' to create directories."

    def __init__(self, base_dir: Path | None = None) -> None:
        self._base_dir = base_dir

    def _resolve(self, path_str: str) -> Path:
        """Resolve a path, making relative paths absolute against base_dir."""
        p = Path(path_str)
        if not p.is_absolute() and self._base_dir:
            return self._base_dir / p
        return p

    def _parameters_schema(self) -> dict[str, Any]:
        return {
            "type": "object",
            "properties": {
                "action": {
                    "type": "string",
                    "enum": [
                        "read",
                        "write",
                        "append",
                        "edit",
                        "mkdir",
                        "list",
                        "exists",
                        "delete",
                    ],
                    "description": "The file operation to perform.",
                },
                "path": {
                    "type": "string",
                    "description": "File or directory path.",
                },
                "content": {
                    "type": "string",
                    "description": "Content to write (for write/append actions).",
                },
                "old_text": {
                    "type": "string",
                    "description": "Text to find (for edit action — search and replace).",
                },
                "new_text": {
                    "type": "string",
                    "description": "Replacement text (for edit action — search and replace).",
                },
            },
            "required": ["action", "path"],
        }

    async def execute(self, **kwargs: object) -> ToolResult:
        """Execute a file operation.

        Args:
            action: read | write | append | edit | mkdir | list | exists | delete
            path: File or directory path.
            content: Content to write (for write/append).
            old_text: Text to find (for edit).
            new_text: Replacement text (for edit).
        """
        action = str(kwargs.get("action", ""))
        path_str = str(kwargs.get("path", ""))

        if not action or not path_str:
            return ToolResult(
                tool_name=self.name,
                success=False,
                error="Both 'action' and 'path' are required",
            )

        path = self._resolve(path_str)

        try:
            if action == "read":
                return await self._read(path)
            elif action == "write":
                content = str(kwargs.get("content", ""))
                return await self._write(path, content)
            elif action == "append":
                content = str(kwargs.get("content", ""))
                return await self._append(path, content)
            elif action == "edit":
                old_text = str(kwargs.get("old_text", ""))
                new_text = str(kwargs.get("new_text", ""))
                return await self._edit(path, old_text, new_text)
            elif action == "mkdir":
                return await self._mkdir(path)
            elif action == "list":
                return await self._list(path)
            elif action == "exists":
                return await self._exists(path)
            elif action == "delete":
                return await self._delete(path)
            else:
                return ToolResult(
                    tool_name=self.name,
                    success=False,
                    error=f"Unknown action: {action}",
                )
        except OSError as exc:
            return ToolResult(tool_name=self.name, success=False, error=str(exc))

    async def _read(self, path: Path) -> ToolResult:
        if not path.exists():
            return ToolResult(tool_name=self.name, success=False, error=f"File not found: {path}")
        content = path.read_text(encoding="utf-8")
        return ToolResult(tool_name=self.name, success=True, output=content)

    async def _write(self, path: Path, content: str) -> ToolResult:
        path.parent.mkdir(parents=True, exist_ok=True)
        # Atomic write
        tmp_fd, tmp_path = tempfile.mkstemp(dir=str(path.parent), suffix=".tmp")
        try:
            with os.fdopen(tmp_fd, "w", encoding="utf-8") as f:
                f.write(content)
            Path(tmp_path).replace(path)
        except Exception:
            Path(tmp_path).unlink(missing_ok=True)
            raise
        return ToolResult(tool_name=self.name, success=True, output=f"Written: {path}")

    async def _append(self, path: Path, content: str) -> ToolResult:
        path.parent.mkdir(parents=True, exist_ok=True)
        with open(path, "a", encoding="utf-8") as f:
            f.write(content)
        return ToolResult(tool_name=self.name, success=True, output=f"Appended to: {path}")

    async def _edit(self, path: Path, old_text: str, new_text: str) -> ToolResult:
        """Search-and-replace edit on a file."""
        if not path.exists():
            return ToolResult(tool_name=self.name, success=False, error=f"File not found: {path}")
        if not old_text:
            return ToolResult(
                tool_name=self.name, success=False, error="old_text is required for edit"
            )

        content = path.read_text(encoding="utf-8")
        if old_text not in content:
            return ToolResult(
                tool_name=self.name,
                success=False,
                error=f"Text not found in {path}. The exact text to replace was not found.",
            )

        new_content = content.replace(old_text, new_text, 1)

        # Atomic write
        tmp_fd, tmp_path = tempfile.mkstemp(dir=str(path.parent), suffix=".tmp")
        try:
            with os.fdopen(tmp_fd, "w", encoding="utf-8") as f:
                f.write(new_content)
            Path(tmp_path).replace(path)
        except Exception:
            Path(tmp_path).unlink(missing_ok=True)
            raise

        return ToolResult(tool_name=self.name, success=True, output=f"Edited: {path}")

    async def _mkdir(self, path: Path) -> ToolResult:
        path.mkdir(parents=True, exist_ok=True)
        return ToolResult(tool_name=self.name, success=True, output=f"Created: {path}")

    async def _list(self, path: Path) -> ToolResult:
        if not path.exists():
            return ToolResult(tool_name=self.name, success=False, error=f"Not found: {path}")
        if path.is_file():
            return ToolResult(tool_name=self.name, success=True, output=str(path))
        entries = sorted(p.name for p in path.iterdir())
        return ToolResult(tool_name=self.name, success=True, output="\n".join(entries))

    async def _exists(self, path: Path) -> ToolResult:
        exists = path.exists()
        return ToolResult(tool_name=self.name, success=True, output=str(exists))

    async def _delete(self, path: Path) -> ToolResult:
        if not path.exists():
            return ToolResult(tool_name=self.name, success=True, output="Already deleted")
        if path.is_file():
            path.unlink()
        else:
            import shutil

            shutil.rmtree(path)
        return ToolResult(tool_name=self.name, success=True, output=f"Deleted: {path}")
